﻿/////////////////////////////////////////////////////////////////////////////////
//
// Adobe(R) Photoshop(R) filter host for .NET
// http://psfilterhost.codeplex.com/
//
// This software is provided under the Microsoft Public License:
//   Copyright (C) 2012-2013 Nicholas Hayes
// 
// See LICENSE.txt for complete licensing and attribution information.
//
/////////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.Serialization;
using System.Security.Permissions;

namespace PSFilterHostDll
{
    /// <summary>
    /// The class that encapsulates the filter's global parameter data.
    /// </summary>
    /// <threadsafety static="true" instance="false" />
    [Serializable]
    internal sealed class GlobalParameters : ISerializable
    {
        internal enum DataStorageMethod
        {
            HandleSuite,
            OTOFHandle,
            RawBytes
        }

        private byte[] parameterDataBytes;
        private DataStorageMethod parameterDataStorageMethod;
        private byte[] pluginDataBytes;
        private DataStorageMethod pluginDataStorageMethod;

        /// <summary>
        /// Gets the parameter data bytes.
        /// </summary>
        /// <returns>The parameter data bytes.</returns>
        internal byte[] GetParameterDataBytes()
        {
            return parameterDataBytes;
        }

        /// <summary>
        /// Sets the parameter data bytes.
        /// </summary>
        /// <param name="data">The data.</param>
        internal void SetParameterDataBytes(byte[] data)
        {
            parameterDataBytes = data;
        }

        /// <summary>
        /// Gets or sets the storage method of the parameter data.
        /// </summary>
        /// <value>
        /// The parameter data storage method.
        /// </value>
        internal DataStorageMethod ParameterDataStorageMethod
        {
            get
            {
                return parameterDataStorageMethod;
            }
            set
            {
                parameterDataStorageMethod = value;
            }
        }

        /// <summary>
        /// Gets the plug-in data bytes.
        /// </summary>
        /// <returns>The plug-in data bytes.</returns>
        internal byte[] GetPluginDataBytes()
        {
            return pluginDataBytes;
        }

        /// <summary>
        /// Sets the plug-in data bytes.
        /// </summary>
        /// <param name="data">The data.</param>
        internal void SetPluginDataBytes(byte[] data)
        {
            pluginDataBytes = data;
        }

        /// <summary>
        /// Gets or sets the storage method of the plug-in data.
        /// </summary>
        /// <value>
        /// The plug-in data storage method.
        /// </value>
        internal DataStorageMethod PluginDataStorageMethod
        {
            get
            {
                return pluginDataStorageMethod;
            }
            set
            {
                pluginDataStorageMethod = value;
            }
        }
       
        
        /// <summary>
        /// Initializes a new instance of the <see cref="GlobalParameters"/> class.
        /// </summary>
        internal GlobalParameters()
        {
            this.parameterDataBytes = null;
            this.parameterDataStorageMethod = DataStorageMethod.HandleSuite;
            this.pluginDataBytes = null;
            this.pluginDataStorageMethod = DataStorageMethod.HandleSuite;
        }
        private GlobalParameters(SerializationInfo info, StreamingContext context)
        {
            if (info == null)
                throw new ArgumentNullException("info");

            this.parameterDataBytes = (byte[])info.GetValue("parameterDataBytes", typeof(byte[]));
            this.parameterDataStorageMethod = (DataStorageMethod)info.GetValue("parameterDataStorageMethod", typeof(DataStorageMethod));

            this.pluginDataBytes = (byte[])info.GetValue("pluginDataBytes", typeof(byte[]));
            this.pluginDataStorageMethod = (DataStorageMethod)info.GetValue("pluginDataStorageMethod", typeof(DataStorageMethod));
        }
        /// <summary>
        /// Populates a <see cref="T:System.Runtime.Serialization.SerializationInfo"/> with the data needed to serialize the target object.
        /// </summary>
        /// <param name="info">The <see cref="T:System.Runtime.Serialization.SerializationInfo"/> to populate with data.</param>
        /// <param name="context">The destination (see <see cref="T:System.Runtime.Serialization.StreamingContext"/>) for this serialization.</param>
        /// <exception cref="T:System.Security.SecurityException">
        /// The caller does not have the required permission.
        ///   </exception>
        /// <exception cref="System.ArgumentNullException"><paramref name="info"/> is null.</exception>
        [SecurityPermission(SecurityAction.Demand, Flags = SecurityPermissionFlag.SerializationFormatter)]
        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            if (info == null)
                throw new ArgumentNullException("info", "info is null.");

            info.AddValue("parameterDataBytes", this.parameterDataBytes, typeof(byte[]));
            info.AddValue("parameterDataStorageMethod", this.parameterDataStorageMethod, typeof(DataStorageMethod));

            info.AddValue("pluginDataBytes", this.pluginDataBytes, typeof(byte[]));
            info.AddValue("pluginDataStorageMethod", this.pluginDataStorageMethod, typeof(DataStorageMethod));
        }
    }
}
